// Copyright 2015-present 650 Industries. All rights reserved.

#import <AVFoundation/AVFoundation.h>

#import <UMCore/UMDefines.h>

#import <EXAV/EXAudioSessionManager.h>

#import <EXAV/EXAV.h>

NSString * const EXAudioSessionManagerErrorDomain = @"EXAudioSessionManager";

@interface EXAudioSessionManager ()

@property (nonatomic, assign) BOOL sessionIsActive;
@property (nonatomic, strong) NSString *activeCategory;
@property (nonatomic, assign) AVAudioSessionCategoryOptions activeOptions;

@property (nonatomic, weak) id activeScopedModule;
@property (nonatomic, strong) NSMapTable<NSString *, id> *allModules;
@property (nonatomic, strong) NSMutableSet<NSString *> *activeModules;
@property (nonatomic, strong) NSMutableDictionary<NSString *, NSString *> *moduleCategory;
@property (nonatomic, strong) NSMutableDictionary<NSString *, NSNumber *> *moduleCategoryOptions;

@end

@implementation EXAudioSessionManager

UM_REGISTER_SINGLETON_MODULE(AudioSessionManager);

- (instancetype)init
{
  if ((self = [super init])) {
    _sessionIsActive = NO;
    _allModules = [NSMapTable strongToWeakObjectsMapTable];
    _activeModules = [[NSMutableSet alloc] init];
    _moduleCategory = [[NSMutableDictionary alloc] init];
    _moduleCategoryOptions = [[NSMutableDictionary alloc] init];

    AVAudioSession *session = [AVAudioSession sharedInstance];
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(handleAudioSessionInterruption:)
                                                 name:AVAudioSessionInterruptionNotification
                                               object:session];
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(handleMediaServicesReset:)
                                                 name:AVAudioSessionMediaServicesWereResetNotification
                                               object:session];
  }
  return self;
}

- (void)dealloc
{
  [[NSNotificationCenter defaultCenter] removeObserver:self];
}

- (void)handleAudioSessionInterruption:(NSNotification *)notification
{
  if (_activeScopedModule) {
    [_activeScopedModule handleAudioSessionInterruption:notification];
  }
}

- (void)handleMediaServicesReset:(NSNotification *)notification
{
  for (NSString *experienceId in _allModules.keyEnumerator) {
    [[_allModules objectForKey:experienceId] handleMediaServicesReset:notification];
  }
}

#pragma mark - EXAVScopedModuleDelegate

+ (NSString *)getExperienceIdFromScopedModule:(id)scopedModule
{
  if ([scopedModule respondsToSelector:@selector(experienceId)]) {
    return [scopedModule experienceId];
  }
  return nil;
}

- (BOOL)isActiveForScopedModule:(id)scopedModule {
  return _activeScopedModule == scopedModule;
}

- (NSString *)activeCategory
{
  return [[AVAudioSession sharedInstance] category];
}

- (AVAudioSessionCategoryOptions)activeCategoryOptions
{
  return [[AVAudioSession sharedInstance] categoryOptions];
}

- (NSError *)setActive:(BOOL)active forScopedModule:(id)scopedModule {
  NSString *experienceId = [EXAudioSessionManager getExperienceIdFromScopedModule:scopedModule];
  if (!experienceId) {
    return [self _getNoExperienceIdError];
  }

  BOOL overridenActive = [_activeModules containsObject:experienceId];

  if (active) {
    [_activeModules addObject:experienceId];
  } else {
    [_activeModules removeObject:experienceId];
  }

  if (_activeScopedModule == scopedModule && active != overridenActive) {
    return [self _updateSessionConfigurationForScopedModule:scopedModule];
  }
  return nil;
}

- (NSError *)setCategory:(NSString *)category withOptions:(AVAudioSessionCategoryOptions)options forScopedModule:(id)scopedModule {
  NSString *experienceId = [EXAudioSessionManager getExperienceIdFromScopedModule:scopedModule];
  if (!experienceId) {
    return [self _getNoExperienceIdError];
  }

  NSString *overridenCategory = [_moduleCategory objectForKey:experienceId];
  NSNumber *overridenOptions = [_moduleCategoryOptions objectForKey:experienceId];

  [_moduleCategory setObject:category forKey:experienceId];
  [_moduleCategoryOptions setObject:@(options) forKey:experienceId];

  BOOL oldSettingsWerePresent = overridenCategory && overridenOptions;
  BOOL newSettingsAreDifferent = !oldSettingsWerePresent || ![overridenCategory isEqualToString:category] || ([overridenOptions unsignedIntegerValue] != options);

  if (_activeScopedModule == scopedModule && newSettingsAreDifferent) {
    return [self _updateSessionConfigurationForScopedModule:scopedModule];
  }
  return nil;
}

- (void)scopedModuleDidBackground:(id)scopedModule
{
  if (_activeScopedModule == scopedModule) {
    _activeScopedModule = nil;
  }

  // Any possible failures are silent
  [self _updateSessionConfigurationForScopedModule:_activeScopedModule];
}

- (void)scopedModuleDidForeground:(id)scopedModule
{
  _activeScopedModule = scopedModule;
  NSString *experienceId = [EXAudioSessionManager getExperienceIdFromScopedModule:scopedModule];
  @synchronized (_allModules) {
    [_allModules setObject:scopedModule forKey:experienceId];
  }

  // Any possible failures are silent
  [self _updateSessionConfigurationForScopedModule:_activeScopedModule];
}

- (void)scopedModuleWillDeallocate:(id)scopedModule
{
  // In case of reloading we won't get a notification about backgrounding,
  // but still we would like to start afresh with an inactive audio session.
  if (scopedModule == _activeScopedModule) {
    [self scopedModuleDidBackground:scopedModule];
  }

  NSString *experienceId = [EXAudioSessionManager getExperienceIdFromScopedModule:scopedModule];
  @synchronized (_allModules) {
    if ([_allModules objectForKey:experienceId] == scopedModule) {
      [_allModules removeObjectForKey:experienceId];
    }
  }
  [_activeModules removeObject:experienceId];
  [_moduleCategory removeObjectForKey:experienceId];
  [_moduleCategoryOptions removeObjectForKey:experienceId];
}

# pragma mark - Utilities

- (NSError *)_updateSessionConfigurationForScopedModule:(id)scopedModule
{
  NSString *experienceId = [EXAudioSessionManager getExperienceIdFromScopedModule:scopedModule];

  AVAudioSession *session = [AVAudioSession sharedInstance];
  NSError *error;

  BOOL shouldBeActive = [_activeModules containsObject:experienceId];
  NSString *category = [self _getRequestedCategoryForScopedModule:scopedModule];
  AVAudioSessionCategoryOptions options = [self _getCategoryOptionsForScopedModule:scopedModule];

  // If the session ought to be deactivated let's deactivate it and then configure.
  // And if the session should be activated, let's configure it first!

  if (!shouldBeActive && _sessionIsActive) {
    [session setActive:NO error:&error];
    if (!error) {
      _sessionIsActive = NO;
    }
  }

  if (error) {
    return error;
  }

  if (!_activeCategory || ![category isEqualToString:_activeCategory] || options != _activeOptions) {
    [session setCategory:category withOptions:options error:&error];
    if (!error) {
      _activeOptions = options;
      _activeCategory = category;
    }
  }

  if (error) {
    return error;
  }

  if (shouldBeActive && !_sessionIsActive) {
    [session setActive:YES error:&error];
    if (!error) {
      _sessionIsActive = YES;
    }
  }

  if (error) {
    return error;
  }

  return nil;
}

- (NSString *)_getRequestedCategoryForScopedModule:(id)scopedModule
{
  NSString *experienceId = [EXAudioSessionManager getExperienceIdFromScopedModule:scopedModule];
  NSString *category = [_moduleCategory objectForKey:experienceId];
  if (category) {
    return category;
  }
  return AVAudioSessionCategorySoloAmbient;
}

- (AVAudioSessionCategoryOptions)_getCategoryOptionsForScopedModule:(id)scopedModule
{
  NSString *experienceId = [EXAudioSessionManager getExperienceIdFromScopedModule:scopedModule];
  NSNumber *categoryOptions = [_moduleCategoryOptions objectForKey:experienceId];
  if (categoryOptions) {
    return [categoryOptions unsignedIntegerValue];
  }
  return 0;
}

- (NSError *)_getNoExperienceIdError
{
  return [NSError errorWithDomain:EXAudioSessionManagerErrorDomain
                             code:EXAudioSessionManagerErrorCodeNoExperienceId
                         userInfo:@{
                                    NSLocalizedDescriptionKey: @"Experience requesting Audio Session changes has no ID"
                                    }];
}

@end

